interface markerInfo {
    lat: number;
    lng: number;
    details: string;
    name: string;
    status: string;
    unit: string;
    visible: boolean;
    type: string;
    value: number;
}

interface markerChanges {
    removed: markerInfo[];
    added: markerInfo[];
    updated: markerInfo[];
}

class GoogleMapsHelper {
    markers: Record<string, google.maps.Marker> = {};
    lines: Record<string, google.maps.Polyline[]> = {};
    eraseNormalLines = false;
    lastEraseNormalLines = false;
    map: google.maps.Map;
    cluster: MarkerClusterer;
    filterInfo: Record<string, string[]>;
    markersInit = false;

    setMapZoomCallbacks() {
        google.maps.event.addListener(this.map, "zoom_changed", () => {
            var zoomLevel = this.map.getZoom();

            if (zoomLevel < 14) {
                this.eraseNormalLines = true;
            } else {
                this.eraseNormalLines = false;
            }
        });
    }

    initiateMap(markerChanges: markerChanges, linesInfo, map: google.maps.Map) {
        this.map = map;
        const clusterOptions = {
            gridSize: 50,
            zoomOnClick: true,
        };

        this.cluster = new MarkerClusterer(this.map, [], clusterOptions);
        this.setMarkers(markerChanges);
        this.setLinesOfMarkers(linesInfo);
        this.setMapZoomCallbacks();
        this.recreateCluster();
    }

    setMarkers(markerChanges: markerChanges) {
        this.deleteAllMarkers();
        this.markers = {};

        var arrAdd = markerChanges.added;

        for (var key in arrAdd) {
            var icon = this.getIconFromStatusAndType(
                arrAdd[key].status,
                arrAdd[key].type
            );
            this.markers[arrAdd[key].name] = this.setMarker(
                arrAdd[key].lat,
                arrAdd[key].lng,
                icon,
                arrAdd[key].details,
                arrAdd[key].visible,
                arrAdd[key].name,
                arrAdd[key].status,
                arrAdd[key].value,
                arrAdd[key].unit
            );
        }
    }

    setMarker(lat, lng, icon, tooltip, visible, name, status, value, unit) {
        let marker = new google.maps.Marker({
            position: {
                lat: lat,
                lng: lng,
            },
            map: this.map,
            icon: new google.maps.MarkerImage(
                icon,
                null,
                null,
                null,
                new google.maps.Size(32, 32)
            ),
            title:
                "Marker Location: " + lat + "," + lng + " | Status: " + tooltip,
            visible: visible,
        });

        marker["name"] = name;
        marker["visibility"] = visible;
        marker["status"] = status;
        marker["value"] = value;
        marker["unit"] = unit;

        google.maps.event.addListener(marker, "click", () => {
            var icon = marker.getIcon()["url"];
            var status = this.getStatusStringFromStatus(marker["status"]);
            var value = marker["value"];
            var infowindow = new google.maps.InfoWindow({
                content: this.generateNewInfoTable(
                    status,
                    icon,
                    marker["name"],
                    value
                ),
            });
            infowindow.open(this.map, marker);
        });

        return marker;
    }

    setLinesOfMarkers(jsonData) {
        this.lines = {};

        for (var key in jsonData) {
            this.lines[jsonData[key].name] = this.setLinesOfMarker(
                jsonData[key].lines,
                this.markers[jsonData[key].name]
                    ? this.markers[jsonData[key].name].status
                    : null
            );
        }
    }

    getColorFromStatus(status) {
        switch (status) {
            case "1":
                return "828282";

            case "2":
                return "E5791D";

            case "3":
                return "EFD721";

            case "4":
                return "E70000";

            default:
                return "000000";
        }
    }

    getStatusStringFromStatus(status) {
        switch (status) {
            case "1":
                return "OK";

            case "2":
                return "Sem Comunicação";

            case "3":
                return "Alerta";

            case "4":
                return "Crítico";

            default:
                return "";
        }
    }

    getIconFromStatusAndType(status, type) {
        var color = this.getColorFromStatus(status);

        switch (type) {
            case "0":
                return "medidor-" + color + ".png";

            case "1":
                return "bomba-de-agua-" + color + ".png";

            case "2":
                return "tanque-de-agua-" + color + ".png";

            default:
                return "";
        }
    }

    setLinesOfMarker(linesData, status) {
        let arr = [];

        var color = "#" + this.getColorFromStatus(status);

        for (var index in linesData) {
            if (linesData[index].coordinates.length > 1) {
                arr[index] = new google.maps.Polyline({
                    path: linesData[index].coordinates,
                    geodesic: true,
                    strokeColor: color,
                    strokeOpacity: 1.0,
                    strokeWeight: linesData[index].weight
                        ? linesData[index].weight
                        : 2,
                });
                if (status) {
                    arr[index].setMap(this.map);
                }
            }
        }

        return arr;
    }

    updateMarkerIcon(marker, markerJsonData) {
        var icon = marker.getIcon();
        var jsonIcon = this.getIconFromStatusAndType(
            markerJsonData.status,
            markerJsonData.type
        );
        if (icon.url !== jsonIcon) {
            marker.setIcon(
                new google.maps.MarkerImage(
                    jsonIcon,
                    null,
                    null,
                    null,
                    new google.maps.Size(32, 32)
                )
            );
        }
    }

    updateMarkerLineStatus(linesOfMarker, markerJsonData) {
        if (
            linesOfMarker &&
            this.markers[markerJsonData.name].status != markerJsonData.status
        ) {
            this.markers[markerJsonData.name].status = markerJsonData.status;
            for (var i = 0; i < linesOfMarker.length; i++) {
                if (linesOfMarker[i]) {
                    linesOfMarker[i].setOptions({
                        strokeColor:
                            "#" +
                            this.getColorFromStatus(markerJsonData.status),
                    });
                }
            }
        }
    }

    updateMarkerTitle(marker, markerJsonData) {
        var details = marker.getTitle();
        if (details !== markerJsonData.details) {
            marker.setTitle(markerJsonData.details);
        }
    }

    updateMarkerPosition(marker, markerJsonData) {
        var position = marker.getPosition();
        if (
            position.lat() != markerJsonData.lat ||
            position.lng() != markerJsonData.lng
        ) {
            marker.setPosition({
                lat: markerJsonData.lat,
                lng: markerJsonData.lng,
            });
        }
    }

    changeMarkerVisibility(marker, visible) {
        marker.setVisible(visible);
        marker.visibility = visible;

        if (this.lines[marker.name]) {
            if (visible) {
                this.changeMarkerLinesMap(this.lines[marker.name], this.map);
            } else {
                this.changeMarkerLinesMap(this.lines[marker.name], null);
            }
        }
    }

    updateMarkerZoomVisibility(marker, visible) {
        marker.setVisible(visible && marker.visibility);
        if (this.lines[marker.name]) {
            let markerLines = this.lines[marker.name];
            for (var i = 0; i < markerLines.length; i++) {
                if (markerLines[i]) {
                    var weigth = markerLines[i].strokeWeight;
                    if (weigth <= 2) {
                        markerLines[i].setMap(
                            visible && marker.visibility ? this.map : null
                        );
                    }
                }
            }
        }
    }

    updateMarkerValue(markerJsonData) {
        var value = markerJsonData.value + " " + markerJsonData.unit;
        if (value !== this.markers[markerJsonData.name]["value"]) {
            this.markers[markerJsonData.name]["value"] = value;
        }
    }

    updateFiltersInfo(filterInfo: Record<string, string[]>) {
        var applyFiltersAgain =
            JSON.stringify(this.filterInfo) !== JSON.stringify(filterInfo);
        this.filterInfo = filterInfo;

        if (applyFiltersAgain) {
            for (var markerName in this.markers) {
                this.changeMarkerVisibility(
                    this.markers[markerName],
                    !this.checkIfMarkerIsFiltered(
                        this.markers[markerName],
                        this.filterInfo
                    )
                );
            }
        }

        return applyFiltersAgain;
    }

    checkIfMarkerIsFiltered(
        marker: google.maps.Marker,
        flterInfo: Record<string, string[]>
    ) {
        for (var filterProperty in flterInfo) {
            if (
                flterInfo[filterProperty].indexOf(marker[filterProperty]) !== -1
            ) {
                return true;
            }
        }

        return false;
    }

    updateMarkersInfo(
        markerChangesInfo: markerChanges,
        filterInfo: Record<string, string[]>
    ) {
        let filterChanged = false;
        if (this.markers) {
            filterChanged = this.updateFiltersInfo(filterInfo);
        }

        if (!this.markersInit && markerChangesInfo.added) {
            this.markersInit = true;
            markerChangesInfo.added.forEach((newMarker) => {
                if (this.markers[newMarker.name]) {
                    this.removeMarker(newMarker.name);
                }

                var icon = this.getIconFromStatusAndType(
                    newMarker.status,
                    newMarker.type
                );
                this.markers[newMarker.name] = this.setMarker(
                    newMarker.lat,
                    newMarker.lng,
                    icon,
                    newMarker.details,
                    newMarker.visible,
                    newMarker.name,
                    newMarker.status,
                    newMarker.value,
                    newMarker.unit
                );
                if (this.lines[newMarker.name]) {
                    this.changeMarkerLinesMap(
                        this.lines[newMarker.name],
                        this.map
                    );
                    for (
                        var i = 0;
                        i < this.lines[newMarker.name].length;
                        i++
                    ) {
                        if (this.lines[newMarker.name][i]) {
                            this.lines[newMarker.name][i].setOptions({
                                strokeColor:
                                    "#" +
                                    this.getColorFromStatus(newMarker.status),
                            });
                        }
                    }
                }

                let changed = this.checkIfMarkerIsFiltered(
                    this.markers[newMarker.name],
                    this.filterInfo
                );
                this.changeMarkerVisibility(
                    this.markers[newMarker.name],
                    !changed
                );
            });
        }
        let updatedChangedFilter = false;
        if (markerChangesInfo.updated) {
            markerChangesInfo.updated.forEach((markerUpdated) => {
                this.updateMarkerIcon(
                    this.markers[markerUpdated.name],
                    markerUpdated
                );
                this.updateMarkerTitle(
                    this.markers[markerUpdated.name],
                    markerUpdated
                );
                this.updateMarkerPosition(
                    this.markers[markerUpdated.name],
                    markerUpdated
                );
                this.updateMarkerLineStatus(
                    this.lines[markerUpdated.name],
                    markerUpdated
                );
                this.updateMarkerValue(markerUpdated);

                let changed = this.checkIfMarkerIsFiltered(
                    this.markers[markerUpdated.name],
                    this.filterInfo
                );
                this.changeMarkerVisibility(
                    this.markers[markerUpdated.name],
                    !changed
                );

                updatedChangedFilter = changed || updatedChangedFilter;
            });
        }

        if (markerChangesInfo.removed) {
            markerChangesInfo.removed.forEach((removeMarker) => {
                this.removeMarker(removeMarker.name);
            });
        }

        if (filterChanged || updatedChangedFilter || markerChangesInfo.added) {
            this.recreateCluster();
        }

        if (this.eraseNormalLines != this.lastEraseNormalLines) {
            for (var markerName in this.markers) {
                this.updateMarkerZoomVisibility(
                    this.markers[markerName],
                    !this.eraseNormalLines
                );
            }
            this.lastEraseNormalLines = this.eraseNormalLines;
        }
    }

    recreateCluster() {
        this.cluster.clearMarkers();
        var markers = [];
        for (var markerName in this.markers) {
            if (this.markers[markerName].visibility) {
                markers.push(this.markers[markerName]);
            }
        }

        this.cluster.addMarkers(markers);
    }

    removeMarker(markerName: string) {
        if (this.markers[markerName]) {
            this.markers[markerName].setMap(null);
            this.cluster.removeMarker(this.markers[markerName]);
            delete this.markers[markerName];
        }
    }

    changeMarkerLinesMap(markerLines, mapToSet) {
        for (var i = 0; i < markerLines.length; i++) {
            if (markerLines[i]) {
                markerLines[i].setMap(mapToSet);
            }
        }
    }

    deleteAllMarkers() {
        for (var name in this.markers) {
            this.removeMarker(name);
        }
    }

    generateNewInfoTable(status, icon, name, value) {
        var infoTable =
            '<div style=" height: 70px; width: 230px; padding: 5px; border-radius:  5px; background-color: #282d41; color: whitesmoke;">' +
            '<div style="height: 20px;display: flex; border-radius: 5px 5px 0px 0px;justify-content: space-around;background-color: #2d3a48;">' +
            name +
            "</div>" +
            '<div style="display: flex; height: 50px;">' +
            '<div style="width: 50px; height: 100%;">' +
            '<img src="' +
            icon +
            '" style="width: 100%; height: 100%;">' +
            "</div>" +
            '<div style="width: 180px; height: 100%; padding-left: 10px; font-size: smaller;">' +
            ' <div style="display: flex;height: 50%;">' +
            '<div style="height: 100%; width: 30%;">Valor:</div>' +
            '<div style="height: 100%; width: 70%;display: flex;justify-content: space-around;">' +
            value +
            "</div>" +
            "</div>" +
            '<div style="display: flex;height:50%;">' +
            '<div style="height: 100%; width: 30%;">Status:</div>' +
            '<div style="height: 100%; width: 70%;display: flex;justify-content: space-around;">' +
            status +
            "</div>" +
            "</div>" +
            "</div>" +
            "</div>" +
            "</div>";

        return infoTable;
    }
}

var mapsHelper = new GoogleMapsHelper();
